mod Maux logic: 'cl'

incl Mtex

rem 
  \para \bf  Numeric functions.  \end 

fun/2 Max 'Tex_f2_pa_max'
  Max(x,y) = x <- x >= y
  Max(x,y) = y <- x < y

fun/2 Exp 'Tex_f2_pa_exp'
  Exp(x,0) = 1
  Exp(x,y+1) = x*Exp(x,y)

thm Exp_zero
  Exp(x,y) = 0 <-> x = 0 & y > 0
proof 
 ind N; y @ 0; y
  proved
  split* x = 0 \/ x = 0 & y > 0 <-> x = 0.2,0,0
   split x = 0 \/ x = 0 & y > 0.4,0,0,0
    proved
    proved.
   proved...

thm Exp_one
  Exp(x,y) = 1 <-> x = 1 \/ y = 0
proof 
 ind N; y @ 0; y
  proved
  split* x = 1 & (x = 1 \/ y = 0) <-> x = 1.2,0,0
   proved
   proved...

thm Exp_plus
  Exp(x,y+z) = Exp(x,y)*Exp(x,z)
proof 
 ind N; y @ 0; y
  proved
  proved..

thm Exp_minus
  x > 0 & y >= z -> Exp(x,y-z) = Exp(x,y)/Exp(x,z)
proof 
 ind N; y; z @ 0; y
  proved
  case N; z @ 0; z1
   proved
   inst \a z(y >= z -> Exp(x,y-z) = Exp(x,y)/Exp(x,z)); z1 proved....

thm Exp_eq
  x > 1 -> Exp(x,y) = Exp(x,z) <-> y = z
proof 
 ind N; y; z @ 0; y
  use Exp_one.0; x; z proved.
  case N; z @ 0; z1
   proved
   inst \a z(Exp(x,y) = Exp(x,z) <-> y = z).0; z1 proved....

thm Exp_le
  x > 1 -> Exp(x,y) <= Exp(x,z) <-> y <= z
proof 
 ind N; y; z @ 0; y
  use Exp_zero.0; x; z proved.
  case N; z @ 0; z1
   use Exp_zero.0; x; y proved.
   inst \a z(Exp(x,y) <= Exp(x,z) <-> y <= z).0; z1 proved....

thm Exp_lt
  x > 1 -> Exp(x,y) < Exp(x,z) <-> y < z
proof 
 ind N; y; z @ 0; y
  cut z = 0
   proved
   use Exp_zero.0; x; z-1 proved..
  case N; z @ 0; z1
   use Exp_zero.0; x; y proved.
   inst \a z(Exp(x,y) < Exp(x,z) <-> y < z).0; z1 proved....

rem 
  \para \bf  Lists.  \end 

fun L 
  L(0) = 0
  L(v,w) = L(w)+1

thm L_zero
  L(x) = 0 <-> x = 0
proof 
 case Ln; x @ 0; v,w
  proved
  proved..

thm L_conc
  L(x++y) = L(x)+L(y)
proof 
 ind Ln; x @ 0; v,w
  proved
  proved..

fun/2 Sub 'Tex_f2_ln_sub'
  Sub((v,w),0) = v
  Sub((v,w),i+1) = Sub(w,i)

rem 
  \para \bf  Binary trees.  \end 

fun/0 E 'Tex_f0_bt_e'
  E = 0,0

fun/3 Nd 'Tex_f3_bt_nd'
  Nd(x,l,r) = 1,x,l,r

pred Bt 
  Bt(E)
  Bt Nd(x,l,r) <- N(x) & Bt(l) & Bt(r)

pred/2 Inbt 'Tex_p2_bt_in'
  Inbt(x,Nd(y,l,r)) <- x = y
  Inbt(x,Nd(y,l,r)) <- x != y & Inbt(x,l)
  Inbt(x,Nd(y,l,r)) <- x != y & ~Inbt(x,l) & Inbt(x,r)

thm Inbt_e
  ~Inbt(x,E)
proof  proved.

thm Inbt_nd
  Inbt(x,Nd(y,l,r)) <-> x = y \/ Inbt(x,l) \/ Inbt(x,r)
proof 
 cut x = y
  proved
  cut Inbt(x,l)
   proved
   proved...

fun Sz 'Tex_f1_bt_sz'
  Sz(E) = 0
  Sz Nd(x,l,r) = Sz(l)+Sz(r)+1

thm Sz_e
  Bt(t) -> Sz(t) = 0 <-> t = E
proof 
 case Bt; t @ 0; x,l,r
  proved
  proved..

fun Dp 'Tex_f1_bt_dp'
  Dp(E) = 0
  Dp Nd(x,l,r) = Max(Dp(l),Dp(r))+1

thm Dp_zero
  Bt(t) -> Dp(t) = 0 <-> t = E
proof 
 case Bt; t @ 0; x,l,r
  proved
  proved..

thm Dp_sz
  Bt(t) -> Dp(t) <= Sz(t) & Sz(t) < Exp(2,Dp(t))
proof 
 ind Bt; t @ 0; x,l,r
  proved
  case Dich; Dp(r),Dp(l) @ x1,y; x1,y
   use Exp_le.0; 2; Dp(r); Dp(l) proved.
   use Exp_le.0; 2; Dp(l); Dp(r) proved....

pred/2 Subtree 'Tex_p2_bt_subtree'
  Subtree(t1,t2) <- t1 = t2
  Subtree(t1,t2) <- t1 != t2 & t2 = Nd(x2,l2,r2) & Subtree(t1,l2)
  Subtree(t1,t2) <- 
    t1 != t2 & t2 = Nd(x2,l2,r2) & ~Subtree(t1,l2) & Subtree(t1,r2)

thm Subtree_right_e
  Subtree(t,E) <-> t = E
proof 
 cut t = E
  proved
  proved..

thm Subtree_right_nd
  Subtree(t,Nd(x,l,r)) <-> t = Nd(x,l,r) \/ Subtree(t,l) \/ Subtree(t,r)
proof 
 cut t = Nd(x,l,r)
  proved
  cut Subtree(t,l)
   proved
   proved...

pred/2 Isomorphic 'Tex_p2_bt_isomorphic'
  Isomorphic(E,E)
  Isomorphic(Nd(x1,l1,r1),Nd(x2,l2,r2)) <- Isomorphic(l1,l2) & Isomorphic(r1,r2)

rem 
  \para \bf  Braun trees.  \end 

pred/2 Index 
  Index(0,Nd(x,l,r))
  Index(S1(i),Nd(x,l,r)) <- Index(i,l)
  Index(S2(i),Nd(x,l,r)) <- Index(i,r)

lemma Index_e
  ~Index(i,E)
proof 
 case N2; i @ 0; i1; i1
  proved
  proved
  proved..

lemma Index_sz
  Bt(t) -> \e i(i <= Sz(t) & ~Index(i,t))
proof 
 ind Bt; t @ 0; x,l,r
  inst* \e i(i = 0 & ~Index(i,E)); 0 proved.
  case Dich; Sz(l),Sz(r) @ x1,y; x1,y
   weak \e i(i <= Sz(r) & ~Index(i,r))
    eigen \e i(i <= Sz(l) & ~Index(i,l)).0; j
     inst* \e i(i <= Sz(l)+Sz(r)+1 & ~Index(i,Nd(x,l,r))); S1(j) proved...
   weak \e i(i <= Sz(l) & ~Index(i,l))
    eigen \e i(i <= Sz(r) & ~Index(i,r)).0; j
     inst* \e i(i <= Sz(l)+Sz(r)+1 & ~Index(i,Nd(x,l,r))); S2(j) proved......

fun/2 Lookup 'Tex2_bappl1'
  Lookup(Nd(x,l,r),0) = x
  Lookup(Nd(x,l,r),S1(i)) = Lookup(l,i)
  Lookup(Nd(x,l,r),S2(i)) = Lookup(r,i)

pred Brt 
  Brt(t) <-> Bt(t) & \a i(Index(i,t) <-> i < Sz(t))

thm Brt_e
  Brt(E)
proof 
 exp Brt; E
  eigen* \a i(Index(i,E) <-> \f).0; i
   use Index_e; i proved....

lemma Brt_nd_sz
  Brt Nd(x,l,r) -> Sz(l) = Sz(r) \/ Sz(l) = Sz(r)+1
proof 
 cut Sz(l) < Sz(r)
  exp Brt; Nd(x,l,r)
   use Index_sz.0; l
    weak Bt(l); Bt(r)
     eigen \e i(i <= Sz(l) & ~Index(i,l)).0; i
      inst \a i(Index(i,Nd(x,l,r)) <-> i <= Sz(l)+Sz(r)).0; S1(i) proved.....
  cut Sz(l) > Sz(r)+1
   exp Brt; Nd(x,l,r)
    use Index_sz.0; r
     weak Bt(l); Bt(r)
      eigen \e i(i <= Sz(r) & ~Index(i,r)).0; i
       inst \a i(Index(i,Nd(x,l,r)) <-> i <= Sz(l)+Sz(r)).0; S2(i) proved.....
   inv* Sz(l) = Sz(r) proved....

lemma Brt_nd_lsbt
  Brt Nd(x,l,r) -> Brt(l)
proof 
 exp Brt; l
  split* Bt(l) & \a i(Index(i,l) <-> i < Sz(l)).4,0,0,0
   use Brt.2,0,1,0,4,0,0; Nd(x,l,r) proved.
   eigen* \a i(Index(i,l) <-> i < Sz(l)).0; i
    use Brt.2,0,1,0,4,1,0:0,2,1,0; Nd(x,l,r); S1(i)
     weak Index(i,l) <-> 2*i < Sz(l)+Sz(r)
      use Brt_nd_sz; x; l; r
       weak Brt Nd(x,l,r)
        split Sz(l) = Sz(r) \/ Sz(l) = Sz(r)+1.4,0,0,0
         proved
         proved.........

lemma Brt_nd_rsbt
  Brt Nd(x,l,r) -> Brt(r)
proof 
 exp Brt; r
  split* Bt(r) & \a i(Index(i,r) <-> i < Sz(r)).4,0,0,0
   use Brt.2,0,1,0,4,0,0; Nd(x,l,r) proved.
   eigen* \a i(Index(i,r) <-> i < Sz(r)).0; i
    use Brt.2,0,1,0,4,1,0:0,2,1,0; Nd(x,l,r); S2(i)
     weak Index(i,r) <-> 2*i+1 < Sz(l)+Sz(r)
      use Brt_nd_sz; x; l; r
       weak Brt Nd(x,l,r)
        split Sz(l) = Sz(r) \/ Sz(l) = Sz(r)+1.4,0,0,0
         proved
         proved.........

thm Brt_nd
  Brt Nd(x,l,r) <-> (Sz(l) = Sz(r) \/ Sz(l) = Sz(r)+1) & Brt(l) & Brt(r)
proof 
 split* Brt Nd(x,l,r) <-> (Sz(l) = Sz(r) \/ Sz(l) = Sz(r)+1) & Brt(l) & Brt(r).2,
                                                                               0,
                                                                               0
  split* (Sz(l) = Sz(r) \/ Sz(l) = Sz(r)+1) & Brt(l) & Brt(r).4,0,0,0,0
   use Brt_nd_sz; x; l; r
    split Sz(l) = Sz(r) \/ Sz(l) = Sz(r)+1.4,0,0,0
     proved
     proved..
   use Brt_nd_lsbt; x; l; r proved.
   use Brt_nd_rsbt; x; l; r proved..
  exp Brt; Nd(x,l,r)
   split* Bt(l) & Bt(r) & \a i(Index(i,Nd(x,l,r)) <-> i <= Sz(l)+Sz(r)).4,0,0,0,
                                                                        0
    weak Sz(l) = Sz(r) \/ Sz(l) = Sz(r)+1; Brt(r)
     use Brt.2,0,1,0,4,0,0; l proved..
    weak Sz(l) = Sz(r) \/ Sz(l) = Sz(r)+1; Brt(l)
     use Brt.2,0,1,0,4,0,0; r proved..
    eigen* \a i(Index(i,Nd(x,l,r)) <-> i <= Sz(l)+Sz(r)).0; i
     case N2; i @ 0; j; j
      proved
      weak Brt(r); i = S1(j)
       use Brt.2,0,1,0,4,1,0:0,2,1,0; l; j
        weak Brt(l); Index(j,l) <-> j < Sz(l)
         split Sz(l) = Sz(r) \/ Sz(l) = Sz(r)+1.4,0,0,0
          proved
          proved....
      weak Brt(l); i = S2(j)
       use Brt.2,0,1,0,4,1,0:0,2,1,0; r; j
        weak Brt(r); Index(j,r) <-> j < Sz(r)
         split Sz(l) = Sz(r) \/ Sz(l) = Sz(r)+1.4,0,0,0
          proved
          proved..........

thm Brt_index1
  Brt Nd(x,l,r) -> 
  (2*i < Sz(l)+Sz(r) <-> i < Sz(l)) & (2*i+1 < Sz(l)+Sz(r) <-> i < Sz(r))
proof 
 use Brt_nd.2,0,1,0,4,0,0; x; l; r
  weak Brt Nd(x,l,r)
   split Sz(l) = Sz(r) \/ Sz(l) = Sz(r)+1.4,0,0,0
    proved
    proved....

thm Brt_index2
  n = m \/ n = m+1 -> (2*i < n+m <-> i < n) & (2*i+1 < n+m <-> i < m)
proof 
 split n = m \/ n = m+1.4,0,0,0
  proved
  proved..

